// Volumetric SSAO
// Implemented by Tomerk
// Optimized by Ethatron

// ---------------------------------------
// TWEAKABLE VARIABLES.


#undef	TEST_MODE
// testMode. set to 1, you can see the raw ssao

#define LUMINANCE_CONSIDERATION
extern float luminosity_threshold = 0.3;
// comment this line to not take pixel brightness into account

#define N_SAMPLES	9
// number of samples, currently do not change.

extern float aoRadiusMultiplier = 8.0;
// Linearly multiplies the radius of the AO Sampling

extern float aoStrengthMultiplier = 1.0;
// Linearly multiplies the strength of the AO

extern float aoClamp = 0.5;
// The maximum strength of the AO, 0 is max strength, 1 is weakest

extern float ThicknessModel = 100;
// units in space the AO assumes objects' thicknesses are

extern float2 fogParams = float2( 8000, 20500 );

// END OF TWEAKABLE VARIABLES.
// ---------------------------------------


#include "includes/Random.hlsl"
#include "includes/Depth.hlsl"
#define	irangeZ	-oblv_ProjectionDepthRange_MAINPASS.z

#define	rcpres		oblv_ReciprocalResolution_MAINPASS
#define	aspect		(oblv_ReciprocalResolution_MAINPASS.z)
#define	raspect		(1.0 / aspect)
// .x - 1 / width
// .y - 1 / height
// .z - width / height
// .w - width * height
float4 oblv_ReciprocalResolution_MAINPASS;

#define	FOV	oblv_ProjectionFoV_MAINPASS.x
// .x - fovX rad
// .y - fovY rad
// .z - fovX deg
// .w - fovY deg
float4 oblv_ProjectionFoV_MAINPASS;


texture obge_LastRendertarget0_EFFECTPASS;

sampler PassSamplerL = sampler_state
{
	texture = <obge_LastRendertarget0_EFFECTPASS>;

	AddressU = CLAMP;
	AddressV = CLAMP;

	MINFILTER = LINEAR;
	MAGFILTER = LINEAR;
};

struct VSOUT
{
	float4 vertPos : POSITION;
	float2 UVCoord : TEXCOORD0;
};

struct VSIN
{
	float4 vertPos : POSITION0;
	float2 UVCoord : TEXCOORD0;
};

VSOUT FrameVS(VSIN IN)
{
	VSOUT OUT = (VSOUT)0.0f;	// initialize to zero, avoid complaints.

	OUT.vertPos = IN.vertPos;
	OUT.UVCoord = IN.UVCoord;

	return OUT;
}

static const float2 g_InvFocalLen = {
	tan(0.5f * FOV) * raspect,
	tan(0.5f * FOV)
};

static float3 sample_offset[N_SAMPLES] =
{
//#if N_SAMPLES >= 9
	float3(-0.1376476f,  0.2842022f,  -0.9842022f ),
	float3(-0.626618f ,  0.4594115f,  0.1594115f ),
	float3(-0.8903138f, -0.05865424f, -0.55865424f),
	float3( 0.2871419f,  0.8511679f , -0.2511679f ),
	float3(-0.1525251f, -0.3870117f, 0.4870117f  ),
	float3( 0.6978705f, -0.2176773f, 0.8176773f  ),
	float3( 0.7343006f,  0.3774331f,  -0.3774331f  ),
	float3( 0.1408805f, -0.88915f , -0.68915f     ),
	float3(-0.6642616f, -0.543601f, 0.743601f  )
//#endif
};

static float sample_radius[N_SAMPLES] =
{
//#if N_SAMPLES >= 9
	0.948832,
	0.629516,
	0.451554,
	0.439389,
	0.909372,
	0.682344,
	0.5642,
	0.4353,
	0.5130
//#endif
};

float3 getPosition(in float2 uv, in float eye_z) {
	uv = (uv * float2(2.0, -2.0) - float2(1.0, -1.0));
	float3 pos = float3(uv * g_InvFocalLen * eye_z, eye_z);
	return pos;
}

float Occlusion1(VSOUT IN, float rsclxy) {
	float depth = LinearDepth(IN.UVCoord);

	float3 pos = getPosition(IN.UVCoord, depth);
	float3 dx = ddx(pos);
	float3 dy = ddy(pos);
	float3 norm = normalize(cross(dx, dy));
	norm.y *= -1;

	float sample_depth;

	float2 rand_v = rand_2_10( IN.UVCoord );
	float3 rand_vec = float3(rand_v.x, rand_v.y, rand_v.x* rand_v.y);

	   	float3 arrRays[N_SAMPLES] =
   	{
     	float3(-1,-1,-1),
     	float3(-1,-1, 1),
		float3( 1,-1, 1),
     	float3( 1,-1,-1),
     	float3(-1, 1,-1),
     	float3(-1, 1, 1),
		float3( 1, 1, 1),
     	float3( 1, 1,-1),
		float3( 0, 0, 1)
	};
	float ao = 0.0f;
//	[unroll]

	float zscl=(depth*irangeZ);
	if(zscl>10000.0f)zscl=10000.0f;
	zscl=1.0f-zscl/10000.0f; if(zscl<0.1f)zscl=0.1f;

	for (int i = 0; i < 8; i++) {
		float3 rayEnd = normalize( reflect(arrRays[i], rand_vec) + norm*0.3f );
		float rayEndZ = rayEnd.z+(depth* irangeZ);
		float rayEndSceneZ = LinearDepth(IN.UVCoord + 
										float2(rayEnd.x*rsclxy*sample_radius[i]*aoRadiusMultiplier*zscl/640.0f,
                                               rayEnd.y*rsclxy*sample_radius[i]*aoRadiusMultiplier*zscl/480.0f))* irangeZ;

		float maxDist=300.0f;
		if(rayEndZ > rayEndSceneZ)
		{
			if ((depth*irangeZ)<maxDist)maxDist=(depth*irangeZ);
		
			float dist=rayEndZ - rayEndSceneZ;
			if(dist>maxDist)dist=maxDist;
			ao+=(maxDist-dist)/maxDist;
		}
		
				
	}

	ao /= 8.0f;

	float fogNear = min( fogParams.x, 8000 );
	float fogFar = min( fogParams.y, 20500 );
	if (depth * irangeZ >= fogNear )
		ao *= saturate(lerp(1, 0, (depth * irangeZ - fogNear) / (fogFar-fogNear) ));

	ao = 1.0 - ao * aoStrengthMultiplier;

	return ao;
}

float4 Occlusion(VSOUT IN) : COLOR0 {
	float3 sample = tex2D(PassSamplerL, IN.UVCoord).rgb;
	float depth = LinearDepth(IN.UVCoord);

	//[branch]
	if (depth >= 0.99)
		return float4(sample, 1.0);

	float ao=( Occlusion1(IN, 1.0f)	);
		
	return float4(sample, ao*1.5f);
}


float4 Blur(VSOUT IN) : COLOR0 {
	float4 sample = tex2D(PassSamplerL, IN.UVCoord).rgba;
	float ao = sample.a * 4;

	ao += tex2D(PassSamplerL, IN.UVCoord + float2( rcpres.x, 0)).a * 2;
	ao += tex2D(PassSamplerL, IN.UVCoord + float2(-rcpres.x, 0)).a * 2;
	ao += tex2D(PassSamplerL, IN.UVCoord + float2(0,  rcpres.y)).a * 2;
	ao += tex2D(PassSamplerL, IN.UVCoord + float2(0, -rcpres.y)).a * 2;

	ao += tex2D(PassSamplerL, IN.UVCoord + rcpres                ).a;
	ao += tex2D(PassSamplerL, IN.UVCoord - rcpres                ).a;
	ao += tex2D(PassSamplerL, IN.UVCoord + rcpres * float2(1, -1)).a;
	ao += tex2D(PassSamplerL, IN.UVCoord - rcpres * float2(1, -1)).a;

	ao /= 16;
	sample.a = ao;

	return sample;
}

float4 Combine(VSOUT IN) : COLOR0 {
	float4 sample = tex2D(PassSamplerL, IN.UVCoord).rgba;
	float3 color = sample.rgb;
	float ao = sample.a;

#ifdef LUMINANCE_CONSIDERATION
	float luminance = color.r * 0.3 + color.g * 0.59 + color.b * 0.11;
	float white = 1.0;
	float black = 0;

	luminance = clamp(
		max(black, luminance - luminosity_threshold) +
		max(black, luminance - luminosity_threshold) +
		max(black, luminance - luminosity_threshold), 0.0, 1.0);

	ao = lerp(ao, white, luminance);
#endif

	color *= ao;

#ifdef	TEST_MODE
	return ao;
#endif

	return float4(color, 1);
}

technique t0
<
	int group = EFFECTGROUP_PRE;
	int fxclass = EFFECTCLASS_AO;
	int conditions = EFFECTCOND_ZBUFFER;
>
{
	pass {
		VertexShader = compile vs_3_0 FrameVS();
		PixelShader  = compile ps_3_0 Occlusion();
	}

	pass {
		VertexShader = compile vs_3_0 FrameVS();
		PixelShader  = compile ps_3_0 Blur();
	}

	pass {
		VertexShader = compile vs_3_0 FrameVS();
		PixelShader  = compile ps_3_0 Combine();
	}
}
